/**

  Spago4Q - The Business Intelligence Free Platform

  Copyright (C) 2009 Engineering Ingegneria Informatica S.p.A.

  Spago4Q is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  Spago4Q is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

 **/
package it.eng.spago4q.extractors.polarion;

import it.eng.spago.error.EMFUserError;
import it.eng.spago4q.extractors.AbstractExtractor;
import it.eng.spago4q.extractors.bo.GenericItem;
import it.eng.spago4q.extractors.bo.GenericItemInterface;

import java.lang.reflect.Method;
import java.rmi.RemoteException;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Date;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.StringTokenizer;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import com.polarion.alm.ws.client.types.tracker.CustomField;
import com.polarion.alm.ws.client.types.tracker.EnumOptionId;
import com.polarion.alm.ws.client.types.tracker.WorkItem;

public class PolarionWebServiceExtractor extends AbstractExtractor {

	// Datasource parameters
	public static final String URL = "URL";
	public static final String USER = "USER";
	public static final String PASSWORD = "PASSWORD";
	public static final String KEYSTORE = "keyStore";
	public static final String KEYSTORE_PSW = "keyStorePSW";
	
	// Operation parameters
	public static final String QUERY = "QUERY";
	public static final String FIELDS = "FIELDS";
	public static final String CUSTOM_FIELDS = "CUSTOM_FIELDS";
	public static final String QUERY_PARAMETERS = "QUERY_PARAMETERS";
	public static final String PROJECT_ID = "PROJECT_ID";
	public static final String USELASTEXTRACTIONDATE = "USELASTEXTRACTIONDATE";
	
	public static final String DEFAULT_SEPARATOR = ",";
	private static final String FIELD_ID = "id";
	private static final String FIELD_PROJECT = "project";
	
	protected static final String TYPE_REQUIREMENT = "requirement";
	protected static final String TYPE_CHANGE_REQUEST = "changerequest";
	protected static final String TYPE_BUG = "bug";
	protected static final String TYPE_TEST = "test";
	
	private static SimpleDateFormat polarionDateFormatter = new SimpleDateFormat("yyyyMMdd");
	
	private String url;
	private String user;
	private String password;
	private String keystore;
	private String keystorePassword;
	
	private String query;
	private String[] fields = null;
	private String[] customFields = null;
	private Map<String, String> queryParameters = null;
	private Map<String, String> projectIds = null;
	private Boolean useLastExtractionDade = false;
	
	private WorkItemUtil workItemUtil;
	private Map<String, Method> workItemMethods = null;
	
	private Log log = LogFactory.getLog(PolarionWebServiceExtractor.class);
	
	@Override
	protected List<GenericItemInterface> extract() throws EMFUserError {
		List<GenericItemInterface> genericItems = new ArrayList<GenericItemInterface>();
		try {
			
			workItemUtil = new WorkItemUtil(url, user, password);
			buildMethodMap();
			genericItems = runWSQuery();
			
		} catch (RemoteException re) {
			log.error("Extraction Error: Polarion Log in unsuccessful", re);
			re.printStackTrace();
		} catch (Exception e) {
			log.error("Extraction Error: "+e, e);
			e.printStackTrace();
		}
		
		return genericItems;
	}

	private List<GenericItemInterface> runWSQuery() {
		List<GenericItemInterface> genericItems = new ArrayList<GenericItemInterface>();
		
		// check PROJECT_ID parameters
		if (projectIds.isEmpty()){
			log.warn("Nessun progetto impostato per la ricerca!");
			return genericItems;
		}
		
		String execQuery = null;
		String project = null;
		String type = null;
		WorkItem workItem = null;
		long start;
		for (Iterator<String> iterator = projectIds.keySet().iterator(); iterator.hasNext();) {
			project = (String) iterator.next();
			log.info("Estrazione per il progetto: " + project);
			
			// build query
			execQuery = buildQuery(project);
		
			// call Polarion Web Service
			start = System.currentTimeMillis();
			String sort = FIELD_ID;
			log.info("Query: " + execQuery);
			workItemUtil.setFields(fields);
			WorkItem[] items = new WorkItem[0];
			try {
				items = workItemUtil.getWorkItemsByQuery(execQuery, sort);
			} catch (Exception e) {
				log.error("Errore nell'estrazione da Web Service Polarion, estrazione abbandonata! ["+e+"]", e);
				e.printStackTrace();
			}
			log.info("WorkItem  estratti: " + items.length + " in: " + (System.currentTimeMillis() - start) + " ms");
			
			// build generic items list
			start = System.currentTimeMillis();
			GenericItem genericItem = null;
			for (int i = 0; i<items.length; i++) {
				workItem = items[i];
				type = workItemUtil.getWorItemType(workItem);
				if (type == null) {
					continue;
				}
				
				genericItem = createGenericItem(workItem);				
				genericItems.add(genericItem);
				
				// add project field
				genericItem.setValue(FIELD_PROJECT, projectIds.get(project));

				// add WorkItems Type specific fields
				if (type != null && type.equalsIgnoreCase(TYPE_REQUIREMENT)) {
					RequirementsUtil requirementsUtil = new RequirementsUtil(workItemUtil);
					requirementsUtil.setAdditionalFields(workItem, genericItem, project);
				}
				
				if (type != null && type.equalsIgnoreCase(TYPE_BUG)) {
					BugsUtil bugsUtil = new BugsUtil(workItemUtil);
					bugsUtil.setAdditionalFields(workItem, genericItem, project);
				}
				
				// TODO da testare perch al momento non ci sono WI di test collegati a requisiti
				if (type != null && type.equalsIgnoreCase(TYPE_TEST)) {
					TestsUtil testsUtil = new TestsUtil(workItemUtil);
					testsUtil.setAdditionalFields(workItem, genericItem, project);
				}
			}
			
			log.info("Lista GenericItems costruita in: " + (System.currentTimeMillis() - start) + " ms");
		}
		return genericItems;
	}

	private GenericItem createGenericItem(WorkItem workItem) {
		GenericItem genericItem = new GenericItem();
		
		String field;
		Object objectValue = null;
		String value = null;
		Method method;
		
		// set standatd fields values to the generic item
		for (int i = 0; i < fields.length; i++) {
			field = fields[i];
			value = null;
			try {
				method = (Method)workItemMethods.get(field);
				if (method != null) {
					objectValue = method.invoke(workItem);
					value = workItemUtil.getSimpleValue(objectValue);
				} 				
				if (value != null) {
					genericItem.setValue(field, value);
				}
			} catch (Exception e) {
				// TODO: handle exception
				log.warn("Campo: " + field + " non estratto dal WI: " + workItem.toString());
			}
		}
		
		// set custom fields values to the generic item
		for (int i = 0; i < customFields.length; i++) {
			field = customFields[i];
			value = null;
			try {
				CustomField customField = workItemUtil.getCustomField(workItem.getUri(), field);
				if (customField != null && customField.getValue() != null) {
					if (customField.getValue() instanceof EnumOptionId) {
						value = ((EnumOptionId) customField.getValue()).getId();
					} else {
						value = customField.getValue().toString();
					}
				}
				if (value != null) {
					genericItem.setValue(field, value);
				}
			} catch (Exception e) {
				// TODO: handle exception
				log.warn("Campo: " + field + " non estratto dal WI: " + workItem.toString());
			}
		}
		
		return genericItem;
	}

	private String buildQuery(String project) {
		Iterator<String> it;
		String returnQuery = null;
		
		// replace query parameters
		it = queryParameters.keySet().iterator();
		while (it.hasNext()) {
			String key = it.next();
			returnQuery = query.replaceFirst(key, (String)queryParameters.get(key));
		}
		
		if (project != null) {
			returnQuery += " AND (";
			returnQuery += "project.id:"+project;
			returnQuery += ")";
		}
		
		// add time filter (last extraction date)
		Date lastExtraction = null;
		if (useLastExtractionDade){
			try {
				lastExtraction = getLastExtractionDate();
				query += " AND updated:["+polarionDateFormatter.format(lastExtraction)+" TO $today$]";
			} catch (Exception e) {
				log.warn("Unable to get Last Extraction Date");
			}
		}
		
		return returnQuery;
	}

	@Override
	protected void setUp() {
		StringTokenizer st;
		
		long start = System.currentTimeMillis();
		url = readDataSourceParameterValue(URL);
		user = readDataSourceParameterValue(USER);
		password = readDataSourceParameterValue(PASSWORD);
		keystore = readDataSourceParameterValue(KEYSTORE);
		keystorePassword = readDataSourceParameterValue(KEYSTORE_PSW);

		query = readOperationParameterValue(QUERY);		
		useLastExtractionDade = Boolean.valueOf(readOperationParameterValue(USELASTEXTRACTIONDATE));
		
		// build query parameters map
		queryParameters = new HashMap<String, String>();
		String queryParametersParameter = readOperationParameterValue(QUERY_PARAMETERS);
		st = new StringTokenizer(queryParametersParameter, DEFAULT_SEPARATOR);
		while(st.hasMoreTokens()){
			String value = st.nextToken();
			String nome = value.substring(0,value.indexOf('='));
			String valore = value.substring(value.indexOf('=')+1);
			queryParameters.put(nome.trim(), valore.trim());
		}
		
		
		// read and split fields
		String fieldsParameter = readOperationParameterValue(FIELDS);
		st = new StringTokenizer(fieldsParameter, DEFAULT_SEPARATOR);
		fields = new String[st.countTokens()];
		int i = 0;
		while(st.hasMoreTokens()){
			fields[i] = st.nextToken().trim();
			i++;
		}
		
		// read and split custom fields
		String customFieldsParameter = readOperationParameterValue(CUSTOM_FIELDS);
		if (customFieldsParameter != null) {
			st = new StringTokenizer(customFieldsParameter, DEFAULT_SEPARATOR);
			customFields = new String[st.countTokens()];
			i = 0;
			while(st.hasMoreTokens()){
				customFields[i] = st.nextToken().trim();
				i++;
			}
		} else {
			customFields = new String[0];
		}
		
		// build PRojects Map aliases
		// Project Code = Polarion Code
		// TODO - manage the prj list as list of params as in the production version
		projectIds = new HashMap<String, String>();
		List<String> projectParameter = readOperationParameterValueList(PROJECT_ID);
		String project = null;
		String value = null;
		String key = null;
		for (Iterator<String> iterator = projectParameter.iterator(); iterator
				.hasNext();) {
			project = (String) iterator.next();
			value = project.substring(0,project.indexOf('='));
			key = project.substring(project.indexOf('=')+1);
			projectIds.put(key.trim(), value.trim());
		}
		
		if (keystore!=null && keystore.length()>0){
		    System.setProperty("javax.net.ssl.trustStore", keystore);
		    System.setProperty("javax.net.ssl.trustStorePassword", keystorePassword);
		    System.setProperty("javax.net.ssl.keyStore", keystore);
		    System.setProperty("javax.net.ssl.keyStorePassword", keystorePassword);	
		}
	}

	@Override
	protected void tearDown() {
		// TODO Auto-generated method stub
		
	}

	private void buildMethodMap() throws EMFUserError {
		
		workItemMethods = new HashMap<String, Method>();
		Method[] methods = WorkItem.class.getMethods();
		
		// prendo i valori elencati nell'array 'fields' e non la lista dei campi
		// da mappare nell'interfaccia.
		for (int f = 0; f < fields.length; f++) {
			for (int m = 0; m < methods.length; m++) {
				if (methods[m].getName().equalsIgnoreCase("get"+fields[f])) {
					workItemMethods.put(fields[f], methods[m]);
				}
			}
		}
	}
	
	
	


	

}
