/**

  Spago4Q - The Business Intelligence Free Platform

  Copyright (C) 2009 Engineering Ingegneria Informatica S.p.A.

  Spago4Q is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  Spago4Q is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

 **/
package it.eng.spago4q.extractors.polarion;

import java.net.MalformedURLException;
import java.rmi.RemoteException;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Date;
import java.util.GregorianCalendar;
import java.util.List;

import javax.xml.rpc.ServiceException;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import com.polarion.alm.ws.client.WebServiceFactory;
import com.polarion.alm.ws.client.session.SessionWebService;
import com.polarion.alm.ws.client.tracker.TrackerWebService;
import com.polarion.alm.ws.client.types.Text;
import com.polarion.alm.ws.client.types.projects.User;
import com.polarion.alm.ws.client.types.tracker.Category;
import com.polarion.alm.ws.client.types.tracker.Comment;
import com.polarion.alm.ws.client.types.tracker.CustomField;
import com.polarion.alm.ws.client.types.tracker.EnumOptionId;
import com.polarion.alm.ws.client.types.tracker.TimePoint;
import com.polarion.alm.ws.client.types.tracker.WorkItem;

public class WorkItemUtil {

	// WebService per connettersi a Polarion
	private WebServiceFactory factory = null;
	private SessionWebService sessionService;
	private TrackerWebService trackerService;
	
	// Campi da estrarre dai WorkItem
	private String[] fields = null;
	
	public static final String DEFAULT_DATE_FORMAT = "yyyy-MM-dd HH:mm:ss.SSS Z";
	private static SimpleDateFormat dateFormatter = new SimpleDateFormat(DEFAULT_DATE_FORMAT);
	
	private Log log = LogFactory.getLog(WorkItemUtil.class);
	
	public WorkItemUtil(String url,String name,String password) throws Exception {
		beginSession(url, name, password);
	}
	
	private void beginSession(String url, String name, String password) throws Exception {
		try {
			
			factory = new WebServiceFactory(url);
			
			// setup Web Service
			sessionService = factory.getSessionService();
			trackerService = factory.getTrackerService();
			
			sessionService.logIn(name, password);
		
		} catch (MalformedURLException me) {
			log.error("Provided URL is malformed - protocol unknown.");
			throw me;
		} catch (ServiceException se) {
			log.error("Unreachable web services at Polarion server.");
			throw se;
		} catch (RemoteException re) {
			log.error("Log in unsuccessful");
			throw re;
		} catch (Exception ioe) {
			log.error("Properties not loaded.");
			throw ioe;
		}
    }
	
	public void setFields(String[] fields) {
		this.fields = fields;
	}
	
	public WorkItem[] getWorkItemsByQuery(String query,String sort) throws Exception {
		return getWorkItemsByQuery(query, sort, 0);
	}
	
	private WorkItem[] getWorkItemsByQuery(String query,String sort, int retry) throws Exception {
		
		WorkItem[] items = null;
		try {
			items = trackerService.queryWorkItems(query, sort, fields);
			
		} catch (Exception e) {
			retry++;
			
			if (retry<3){
				log.debug("Errore ("+retry+") nel getWorkItemsByQuery ["+e+"].");
				return getWorkItemsByQuery(query, sort, retry);
				
			} else {
				log.error("Errore nel getWorkItemsByQuery ["+e+"].");
				e.printStackTrace();
				throw e;
			}
		}

		if (items==null)
			items= new WorkItem[0];
		
		return items;
	}
	
	/**
	 * Gets the simple value performing these transformation:
	 * <ul><li>EnumOption -> String</li>
	 * <li>User -> String</li>
	 * <li>User[] -> String</li>
	 * <li>Text -> String</li>
	 * <li>Comment -> String</li>
	 * <li>GregorianCalendar -> String</li>
	 * <li>Categories -> String</li>
	 * <li>TimePoint -> String</li></ul>
	 * 
	 * @param value the value
	 * 
	 * @return the simple value
	 */
	public String getSimpleValue(Object value ) {
		String toReturn = null;
		
		// EnumOption -> String
		if (value instanceof EnumOptionId) {
			value = ((EnumOptionId) value).getId();
		
		// User -> String
		} else if (value instanceof User) {
			value = ((User) value).getName();
			
		// User[] -> String
		} else if (value instanceof User[]) {
			String temp = "";
			for (int i = 0; i<((User[]) value).length; i++) {
				temp += ((User[]) value)[i].getName();
			}
			value = temp;
			
    	// Text
		} else if (value instanceof Text) {
			value = ((Text) value).getContent();
			
		// Comment
		} else if (value instanceof Comment) {
			if (((Comment) value).getText()==null)
				return null;
			value = ((Comment) value).getText().getContent();
    		
		// Data
		} else if (value instanceof Date || value instanceof GregorianCalendar) {
			if (value instanceof GregorianCalendar)
				value=((GregorianCalendar)value).getTime();
			else
				value = ((Date)value);
			value = dateFormatter.format(value);
			
		// Categories
		} else if (value instanceof Category[]) {
			Category[] cat = (Category[])value;
			String temp = "";
			for (int i=0; i<cat.length; i++) {
				temp += cat[i].getName();
			}
			value = temp;
		
		// TimePoint
		} else if (value instanceof TimePoint) {
			value = ((TimePoint) value).getName();
		}
		
		if (value != null){
			toReturn = value.toString();
		}
		
		return toReturn;
	}

	protected boolean isCustomField(String field, WorkItem workItem) throws Exception {
		return isCustomField(field, workItem, 0);
	}
	
	private boolean isCustomField(String parameterName, WorkItem workItem, int retry) throws Exception {
		boolean custom = false;
		
		try {
			String[] customFields = trackerService.getCustomFieldKeys(workItem.getUri());
			for (int i = 0; i < customFields.length && !custom; i++) {
				if (parameterName.equalsIgnoreCase(customFields[i])) {
					custom = true;
				}
			}
		
		} catch (RemoteException e) {
			retry++;
			
			if (retry<3){
				log.debug("Errore ("+retry+") nel isCustomField per il campo "+parameterName+" ["+e+"].");
				return isCustomField(parameterName, workItem, retry);
				
			} else {
				log.error("Errore nel isCustomField per il campo "+parameterName+" ["+e+"].");
				e.printStackTrace();
				throw e;
			}
		}
		
		return custom;
	}
	
	protected CustomField getCustomField(String workItemUri, String filed) throws Exception {
		return getCustomField(workItemUri, filed, 0);
	}
	
	private CustomField getCustomField(String workItemUri, String field, int retry) throws Exception {
		CustomField customField = null;
		
		try {
			customField = trackerService.getCustomField(workItemUri, field);
			
		} catch (Exception e) {
			retry++;
			
			if (retry<3){
				log.debug("Errore ("+retry+") nel getCustomField per il campo "+field+" ["+e+"].");
				return getCustomField( workItemUri, field, retry);
				
			} else {
				log.error("Errore nel getCustomField per il campo "+field+" ["+e+"].");
				if (log.isDebugEnabled()) {
					e.printStackTrace();
				}
				throw e;
			}
		}
		
		return customField;
	}
	
	public List<WorkItem> getChildren(WorkItem workItem, String type, String project) {
		List<WorkItem> children = null;
		
		String fatherId = workItem.getId();
		children = new ArrayList<WorkItem>();
		
		String sort = "id";
		String query = "linkedWorkItems:"+fatherId+" AND project.id:"+project;
		if (type!=null)
			query += " AND type:"+type;
		
		try {
			WorkItem[] linkedWorkItems = getWorkItemsByQuery(query, sort);
			
			if (linkedWorkItems!=null) {
				for (int i = 0; i<linkedWorkItems.length; i++) {
					children.add(linkedWorkItems[i]);
				}
			}
		
		} catch (Exception e) {
			log.error("Errore nel recuperare i figli del WorkItem ["+workItem.getId()+"]! "+e);
		}
			
		return children;
	}

	public String getWorItemType(WorkItem workItem) {
		String type = null;
		try {
			type = workItem.getType().getId();
		} catch (Exception e) {
			log.info("Impossibile determinare il tipo del WI");
			if (log.isDebugEnabled()) {
				e.printStackTrace();
			}
		}
		return type;
	}	
	
    public List<WorkItem> getFather(WorkItem workItem, String type, String project) {
		
		List<WorkItem> fathers = new ArrayList<WorkItem>();
		String childId = workItem.getId();
		
		String sort = "id";
		String query = "backlinkedWorkItems:"+childId+" AND project.id:"+project;
		if (type != null)
			query += " AND type:"+type;
		
		try {
			WorkItem[] backlinkedWorkItems = getWorkItemsByQuery(query, sort);
			
			if (backlinkedWorkItems!=null) {
		    	for (int i = 0; i<backlinkedWorkItems.length; i++) {
		    		fathers.add(backlinkedWorkItems[i]);
		    	}
	    	}
			
		} catch (Exception e) {
			log.error("Errore nel recuperare i padri del WorkItem ["+workItem.getId()+"]! " + e);
		}
		
		return fathers;
	}
}
